WAIT_FOR_BLIT = 1
	;;
	;; 8bpp -> 32bpp blitter code
	;;
	;; Copyright (c) 1997-2000 VM Labs, Inc.
	;; All rights reserved.
	;; Confidential and Proprietary Information
	;; of VM Labs, Inc.
	;;

	;;
	;; helper MPE program; runs in 1 or more
	;; MPEs (other than MPE 0, which is running
	;; the main code). This code sits and waits for
	;; requests to copy an 8 bit CLUT
	;; image (in raster order) to
	;; the frame buffer (in pixel order)
	;;
	;; Requests come in over the comm bus.
	;;
	;; the request number is in the
	;; last longword of the vector.
	;; Valid requests are:
	;;
	;; Set up destination output parameters
	;; v[0] == 0x1  SET_DEST_PARMS
	;; v[1] == base address for DMA
	;; v[2] == flags for DMA
	;; v[3] == unused, reserved
	;;
	;; Set up CLUT for clut based modes
	;; v[0] == 0x2 SET_CLUT
	;; v[1] == address of CLUT
	;; v[2] == unused
	;; v[3] == unused
	;;
	;; Set up source output parameters
	;; v[0] == 0x3 SET_SOURCE_PARMS
	;; v[1] == base address (for bilinear DMA)
	;;         or starting address of linear data
	;; v[2] == flags for DMA (we can tell whether it's linear or bilinear from this)
	;; v[3] == width (in bytes) of linear map;
	;;         unused for bilinear DMA
	;;
	;; An actual blit command occupies 2 comm bus packets,
	;; as follows:
	;;
	;;packet 1:
	;; v[0] == 0x4 DO_BLIT
	;; v[1] == (xcount<<16)|xinit
	;; v[2] == (ycount<<16)|yinit
	;; v[3] == ((0.16 shading)<<16)|(mask for u and v)
	;;packet2:
	;; v[0] == uinit: initial value for source X (16.16)
	;; v[1] == vinit: initial value for source Y (16.16)
	;; v[2] == uinc: value to add to U at each pixel (16.16)
	;; v[3] == vinc: value to add to V at each pixel (16.16)
	;;
	;; ASSUMPTION: scaling is uniform in both directions,
	;; so ustep and vstep (the amount to add at each
	;; line) are assumed to be vinc and uinc, respectively
	;;
	
	;; vector alignment sometimes needed
	ALIGN = 15

	.module blit
	
	;; number of pixels to write at a time
	OUTSIZE = 64
	
	;; number of longwords to read at a time
	INPSIZE = 64

	PIXELSPERWORD = 4

	SET_DEST_PARMS = 1
	SET_CLUT = 2
	SET_SOURCE_PARMS = 3
	DO_BLIT = 4

	;; restrict v to 0..1
	UV_TILE = (0<<16)|(14<<12)

	.if defined(USE_MINI_BIOS)
	.segment mblitc
	.else
	.segment blitc
	.endif
	
	.origin 0x20300000
start:
	;; set up (uv) (input) for 8 bit pixels
	;; also set up the CLUT base at the same time

	mv_s	#CLUT,r1
	st_io	r1,clutbase
	st_io	#(0<<28)|(3<<20)|(INPSIZE*4)|UV_TILE,uvctl
	st_io	#inp_buf,uvbase
	st_io	#0,ru
	st_io	#0,rv
	
	;; set up linear pixel modes for 32 bit pixels
.if defined(USE_MINI_BIOS)
	ld_io	intvec1,r20	; get end of jump table
.else
	st_s	#(1<<9),mpectl	
.endif
	bra	wait_for_request
	st_io	#(1<<28)|(4<<20),linpixctl
	nop

	;; request_loop: ACK the last request, then
	;; wait for the next one
request_loop:
.if defined(USE_MINI_BIOS)
	ld_io	intvec1,r20
.endif
	ld_s	last_request_mpe,r4
.if defined(WAIT_FOR_BLIT)
.if defined(USE_MINI_BIOS)
	sub	#16,r20,r0		;; get address of CommSend
	jsr	(r0)			;; send ACK
	sub	r5,r5
	sub_sv	v0,v0
.else
`waitxmit:
	ld_io	commctl,r0
	or	#$2000,r4
	btst	#15,r0
	bra	ne,`waitxmit,nop
	
	st_io	r4,commctl
	sub_sv	v0,v0
	st_io	r0,comminfo
	st_v	v0,commxmit
.endif
.endif
	
	;; wait for a request to come in
wait_for_request:
.if defined(USE_MINI_BIOS)
	;; r20 has end of jump table
	sub	#8,r20,r0		;; get address of CommRecv
	jsr	(r0),nop

.else		
`request_loop:
	ld_s	(commctl),r4
	nop
	btst	#31,r4
	bra	eq,`request_loop,nop

	;; load the request
	ld_v	(commrecv),v0
	bits	#7,>>#16,r4		;; extract id of MPE that sent request
.endif
	st_s	r4,last_request_mpe
		
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Registers for blit code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	;; v0,v1 are scratch registers (destroyed by subroutines)

	;; used (temporarily) to build DMA command
	dmavec = v0
	
	scratchvec = v1
	temp0 = scratchvec[0]
	temp1 = scratchvec[1]
	
	;; v2: current X and Y positions for input (U,V) and output (X,Y)
	all_curs = v2
	u_cur = v2[0]
	v_cur = v2[1]
	x_cur = v2[2]
	y_cur = v2[3]

	;; v3: per-pixel increment values
	all_incs = v3
	u_inc = v3[0]
	v_inc = v3[1]
	x_inc = v3[2]
	y_inc = v3[3]

	;; v4: pixel I/O stuff
	flushbuffer = v4[0]	; subroutine for outputting pixels
	fillbuffer = v4[1]	; subroutine for reading pixels
	pixels_out = v4[2]	; count of pixels output
	uv_mask = v4[3]		; mask for u and v
	
	;; v5: pixel data
	pixelvec = v5

	;; v6: data pointers and counters
	writebuf = v6[0]	; output buffer for pixels
	dmaoutbuf = v6[1]	; output buffer being used by DMA
	outer_cnt = v6[2]	; outer loop count
	inner_cnt = v6[3]	; inner loop count
	
	;; step values for outer loop
	;; re-used for other things in inner loop
	all_steps = v7
	u_step = v7[0]
	v_step = v7[1]
	x_step = v7[2]
	y_step = v7[3]

	; inner loop uses:
	shade = v7[0]
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Rendering code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

check_render:
	cmp	#DO_BLIT,r0
	bra	ne,check_dest,nop

.if defined(USE_MINI_BIOS)
	ld_io	intvec1,r4
	push	v0
	sub	#8,r4		;; get address of CommRecv
	jsr	(r4),nop
	mv_v	v0,v1
	pop	v0
	nop
.else		
`request_loop:
	ld_s	(commctl),r4
	nop
	btst	#31,r4
	bra	eq,`request_loop,nop

	;; load the request
	ld_v	(commrecv),v1
	nop
.endif
	
	;; r1 == (xcount<<16)|x_cur
	;; r2 == (ycount<<16)|y_cur
	;; r3 == ((shade<<16)|mask)

{	push	v1		; save new packet
	and	#$0000ffff,r3,uv_mask
}
	;; r3 is an 8.24 shading factor; if it is negative, then that
	;; says that we want to do a translucent blit
	;; otherwise, add an "ambient light" factor to it (yeah, it's a hack)
{	cmp	#0,r3
	st_s	#0,img_alphas			;; assume no translucency
}
{	bra	gt,`not_translucent
	add	#fix(0.15,24),r3		;; add ambient light
}
	sat	#25,r3
	lsl	#6,r3	;change from 8.24 to 2.30
	
	mv_s	#0,r3				;; no shading
	st_s	#1,img_alphas			;; use translucency
`not_translucent:
{	st_s	r3,shading_factor
	lsl	#16,r1,x_cur
}
	lsl	#16,r2,y_cur
{	lsr	#16,r1,inner_cnt	; assume we're stepping in X first
	mv_s	#1<<16,x_inc
}
{	lsr	#16,r2,outer_cnt
	mv_s	#0,y_inc
}
{	copy	x_inc,y_step
	mv_s	y_inc,x_step
}

	;; now get the next comm bus packet
	pop	v0
	nop
			
	; r0 == uinit: initial value for source X (16.16)
	; r1 == vinit: initial value for source Y (16.16)
	; r2 == uinc: value to add to U at each pixel (16.16)
	; r3 == vinc: value to add to V at each pixel (16.16)

{	copy	r0,u_cur
	mv_s	r2,u_inc
}
{	copy	r1,v_cur
	mv_s	r3,v_inc
}
	mv_s	u_inc,v_step
	mv_s	v_inc,u_step

.if 1
	;; see if we should render in a different order
	;; if outer_cnt >> inner_cnt, then swap
	;; the order of X and Y; this also means swapping the
	;; "step" and "inc" values
	cmp	outer_cnt,inner_cnt
	bra	ge,`order_ok,nop

	push	all_incs
	mv_v	all_steps,all_incs
	pop	all_steps
	nop
	
{	copy	outer_cnt,inner_cnt
	mv_s	inner_cnt,outer_cnt
}
	
`order_ok:
.endif
			
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Arbitrary line blit code
;; Copy pixels from source bitmap (pointed to by (uv))
;; to the destination bitmap (pointed to by (xy))
;;
;; "dest_dmacmd" is set up to point to the destination; this
;;	must be a bilinear pixmap
;;
;; "src_dmacmd" is set up to point to the source; this must
;;      be a linear pixel map, but may be in either
;;      SDRAM or system RAM
;;
;; the destination line may be either horizontal or
;; vertical, but can't be angled
;;
;; parameters:
;; u_cur == initial value for U
;; v_cur == initial value for V
;; u_inc == amount to add to U on each step
;; v_inc == amount to add to V on each step
;;
;; x_cur == initial X value for output DMA (16.16)
;; y_cur == initial Y value for output DMA (16.16)
;; x_inc == step value for output DMA
;; y_inc == step value for output DMA
;; NOTE: exactly one of x_inc or y_inc must be 1.0,
;;       and the other must be 0.0
;;
;; the copy is done in raster order
;; all registers are modified
;;
;; internal register usage:
;;  ru,rv -- point to next input pixel to load
;;  rx -- points to pixel output buffer
;;  rc0 -- counts number of pixels left (total)
;;  rc1 -- counts number of pixels left in this output buffer
;;  pixels_out == number of pixels put into output buffer
;;  flushbuffer == buffer output routine
;;  fillbuffer == input buffer fill routine
;;
;; ASSUMPTIONS: all *_inc values are positive. This will *not*
;; be a good assumption in general; we'll need some special
;; purpose routines to handle negative steps
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.macro	BLITLOOP u_out_of_range,v_out_of_range
	
`outerloop:
	push	all_curs		;; save current (x,y,u,v) position
	push	all_steps		;; save step values for next line
	
	;; rc0 holds the number of pixels on the line
	;; rc1 holds the number of pixels until the next
	;; output buffer flush; this is the lesser of
	;; OUTSIZE and rc0
	
{	cmp	#OUTSIZE,inner_cnt	;; see if OUTSIZE > inner_cnt
	st_io	inner_cnt,rc0
}
{	bra	lt,`use_inner_cnt,nop	;; if not, use inner_cnt for rc1
	st_io	inner_cnt,rc1
}

	st_io	#OUTSIZE,rc1

`use_inner_cnt:
	;; do the initial buffer load
	jsr	(fillbuffer)
	ld_s	shading_factor, shade
	nop
	
`innerloop:
{	jsr	u_out_of_range,(fillbuffer),nop	;; if u went out of range (we know u_inc > 0)
	range	rv
	dec	rc1
	dec	rc0
}
	jsr	v_out_of_range,(fillbuffer),nop	;; if v went out of range (v_inc > 0)

	;; NOTE: when we come back from fillbuffer,
	;; v0[0] contains the CLUT based index
	
{	addr	u_inc,ru
	add	u_inc,u_cur
	ld_pz	(v0[0]),pixelvec	; do the CLUT lookup
}
{	addr	v_inc,rv
	add	v_inc,v_cur
}
{	bra	c1ne,`innerloop		;; keep going until we need to flush
	mul_p	shade,pixelvec,>>#30,pixelvec
}
{	ld_p	(uv),v0			; pre-load next pixel (we'll throw it away if invalid)
	range	ru
}	
{	st_pz	pixelvec,(xy)
	addr	#(1<<16),rx
	add	#1,pixels_out
}
		

	;; flush the current output buffer
	jsr	(flushbuffer)
	push	v0			; save the pixel we just loaded
	nop

	;; and keep looping if there are more pixels to create
	bra	c0ne,`innerloop
	pop	v0
	nop

	;; see if the outer loop is finished yet
	pop	all_steps
{	pop	all_curs		; restore (x,y,u,v) to beginning of line
	sub	#1,outer_cnt
}
	bra	gt,`outerloop		; BRANCH: continue if more lines to do
{	add	x_step,x_cur		; BRANCH DELAY SLOT: update (x,y)
	addm	y_step,y_cur
}
{	add	u_step,u_cur		; BRANCH DELAY SLOT: update (u,v)
	addm	v_step,v_cur
}

	bra	return
	nop
	nop
	.mend
		
blit_entry:

	;;
	;; general entry point for the blit
	;; assumes *_cur and *_inc are set up
	;; and that inner_cnt == number of pixels to draw on this
	;; line
	;;

	mv_s	#linear8_load,fillbuffer
	mv_s	#out_buf1,writebuf	
	mv_s	#out_buf2,dmaoutbuf
	ld_s	img_alphas,r0		;; check for translucent blit
{	st_io	writebuf,xybase
	sub	pixels_out,pixels_out	;; clear count of output pixels
}
	cmp	#0,r0
{	bra	ne,`translucency_ok,nop
	mv_s	#FlushOutputRMW,flushbuffer
}

	mv_s	#FlushOutput,flushbuffer
`translucency_ok:
	
	;; check which blit routine we should
	;; use
	cmp	#0,u_inc
	bra	lt,negative_u
	cmp	#0,v_inc
	bra	ge,blit_uplus_vplus
	bra	blit_uplus_vneg,nop

negative_u:
	bra	ge,blit_uneg_vplus,nop

	;; fall through to blit_uneg_vneg

	negx = modmi
	posx = modge
	
blit_uneg_vneg:
	BLITLOOP negx,negx
blit_uneg_vplus:
	BLITLOOP negx,posx
blit_uplus_vneg:
	BLITLOOP posx,negx
blit_uplus_vplus:
	BLITLOOP posx,posx
	
return:
	;; wait for all DMAs to finish
`wait1:
	ld_io	odmactl,r0
	ld_io	mdmactl,r1
	bits	#4,>>#0,r0
{	bra	ne,`wait1,nop
	bits	#4,>>#0,r1
}
	bra	ne,`wait1,nop

	bra	request_loop,nop


check_dest:
{	cmp	#SET_DEST_PARMS,r0
	st_v	v0,DEBUG_LAST_CMD
}
	bra	ne,check_source,nop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Set destination output flags
;; sets up DMA flags and base address
;; also modifies (xy) addressing based upon the dma_flags
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	st_s	r1,dest_baseaddr
	st_s	r2,dest_flags

	;; set up (xy) (output) for 16 or 32 bit pixels depending
	;; upon the transfer type
	bits	#3,>>#4,r2		; isolate transfer type
	cmp	#6,r2
{	bra	le,`gotpixtype,nop
	lsl	#20,r2
}
	mv_s	#(5<<20),r2		; assume pixel type 5
`gotpixtype:
	or	#(1<<28)|OUTSIZE,r2
	st_io	r2,xyctl
	st_io	#0,rx
	st_io	#0,ry
	
	bra	return,nop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Set source input flags
;; sets up DMA flags and base address
;; also modifies (xy) addressing based upon the dma_flags
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
check_source:
	cmp	#SET_SOURCE_PARMS,r0
	bra	ne,check_clut,nop
{	st_s	r1,src_baseaddr
	bset	#13,r2			; make sure read flag is set
}
	st_s	r2,src_flags
	st_s	r3,src_width
	
	;; set up (uv) (input) for appropriate pixel type
	;; upon the transfer type
{	bits	#3,>>#4,r2		; isolate transfer type
	mv_s	#3,r1			; assume 8bpp CLUT mode
}
{	st_s	#1,linear_flag
	bra	eq,`gotpixtype,nop	; if 0, then linear transfer
}
{	st_s	#0,linear_flag		; otherwise, bilinear
	cmp	#6,r2			; modes 6 and below map directly
}
{	mv_s	r2,r1
	bra	le,`gotpixtype,nop
}
	mv_s	#5,r1			; assume pixel type 5
	
`gotpixtype:
	lsl	#20,r1
	or	#(1<<28)|(INPSIZE*4)|UV_TILE,r1
	st_io	r1,uvctl

	bra	return,nop
		
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; CLUT loading code
;; loads a new CLUT to use for future blits
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	; register usage: v0-v4 used for converting YCrCb->RGB
	; r28 points at the CLUT
	; r29 is a flag indicating whether any pixels have alpha values
	; or not (unused for now)
	
check_clut:
	cmp	#SET_CLUT,r0
	bra	ne,return,nop

	;; dma in the CLUT data
	jsr	h_dma_read
{	copy	r1,r0
	mv_s	#CLUT,r1
}
	mv_s	#256*4,r2	; 256 CLUT entries

	bra	return,nop

	;;
	;; subroutine -- get a new buffer full of pixels
	;; linear DMA version -- assumes 8bpp linear
	;; data
	;;
	;; parameters:
	;;
	
	new_u = scratchvec[0]
	new_v = scratchvec[1]
	new_u_offset = scratchvec[2]	

linear8_load:
	push	scratchvec,rz
{	ld_io	rc0,r0		; get number of pixels remaining
	copy	u_cur,new_u
}
	ld_s	src_width,r1	; get width of buffer
{	ld_s	src_baseaddr,r2	; get starting address of buffer
	mul	u_inc,r0,>>#16,r0
}
	;; if u_inc < 0, then make new_u = u_cur + u_inc*
{	cmp	#0,u_inc
}
{	bra	ge,`new_u_ok,nop
	mv_s	u_cur,new_u
}
	
{	addm	r0,new_u
	abs	r0
}
`new_u_ok:
	
	;;
	;; figure out how many pixels to get
	;; this is roughly u_inc * pixels_left
	;; but should be no more than INPSIZE long words,
	;; and no less than 4 long words (one vector)
	;; the DMA read routines below will take care of the
	;; minimum, so we only need to worry about the
	;; maximum
	;;
	;; FIXME: we actually should also take v_inc into
	;; account, since there's no point in crossing
	;; a line boundary!!!
	
	lsr	#16,v_cur,r3			;; find floor(v_cur)
	and	uv_mask,r3			;; make it tile
	;; calculate address of first pixel
{	mul	r3,r1,>>#0,r1			;; find floor(v_cur)*src_width
	lsr	#16,new_u,r3			;; find floor(u_cur)
}
	and	uv_mask,r3			;; make it tile
	add	r2,r3				;; r3 now has base + floor(u_cur)
	add	r1,r3				;; r3 now has starting address

	;; r3 contains the (byte) address of our source pixel
	;; however, we can only load on scalar boundaries (vector boundaries
	;; for system RAM, because of the other bus bug)
	;; so we will have to load [r3 - (r3&(~ALIGN))] extra pixels,
	;; and set ru to that offset
	and	#~ALIGN,r3,r2			;; r2 now contains the "real" address to load
	sub	r2,r3				;; r3 contains extra bytes to load

{	addm	r3,r0				;; r0 contains bytes to load
	lsl	#16,r3,new_u_offset		;; new_u contains offset in 16.16 format
}
{	add	#ALIGN,r0			;; round size to next (ALIGN+1) bytes
}
	and	#~ALIGN,r0,r3			;; r3 contains rounded size
{	mv_s	#(ALIGN+1),r0			;; if we get 0, load (ALIGN+1) bytes
	bra	eq,`size_ok,nop
}
	;; otherwise, check for a transfer that's too big
`nonzero_bytes:
{	cmp	#INPSIZE*4,r3			;; is this too many pixels?
	mv_s	r3,r0
}
	bra	le,`size_ok,nop
	mv_s	#INPSIZE*4,r0			;; settle for this many pixels
`size_ok:
	;; size is in r0
	;; starting address is in r2
	
	;; FIXME: we really should do the DMA "by hand"
	;; because we could then do the stuff below
	;; in parallel (plus, h_dma_read has a lot
	;; of overhead for such a simple transfer)

	;; set up cache info
	;; first, find fractional parts of u and v
	;; and add the appropriate offset to u to adjust
	;; for re-alignment
	mv_s	#$0000ffff,r3
	and	v_cur,r3,new_v
{	and	u_cur,r3
	mvr	new_v,rv
}
	mvr	r3,ru
{	addr	new_u_offset,ru
	lsr	#16,new_u_offset	; turn new_u_offset back into an integer
}
`pixel_count_ok:	
	;; calculate new uvrange based upon the number
	;; of pixels loaded
	;; we also want to make sure that we don't
	;; go past the edge of the image, so
	;; limit the number of active pixels to imgwidth - floor(u_cur)
	;; however, note that new_u_offset pixels are
	;; loaded at the beginning of the line (and do nothing
	;; for us)
	;; r0 has the number of pixels to load
{	ld_s	src_width,r5
	lsr	#16,new_u,r4		; r4 == floor(u_cur)
}
	and	uv_mask,r4	; now r4 == floor(y_cur)
	sub	r4,r5		; r5 == #pixels left on line
	add	new_u_offset,r5	; include the "dead" pixels at start of line
	cmp	r0,r5		; if (r5 - r0) >= 0, use r0 for uvrange
{	bra	ge,`setuvrange,nop
	lsl	#16,r0,r3
}

	;; we come here if we want to limit the uvrange to "r5" pixels
	lsl	#16,r5,r3
`setuvrange:
	or	#1,r3
	st_io	r3,uvrange

	;; now do the transfer
	mv_s	#inp_buf,r1
	jsr	h_dma_read
{	copy	r0,r2
	mv_s	r2,r0
}
	nop
	
	pop	scratchvec,rz
	ld_p	(uv),v0			;; load new pixel for the calling code
	nop
	cmp	#$20100000,v0[0]
	bra	lt,`bad_ld_p,nop
	rts
	nop
	range	ru			;; make sure the modge bits are clear
`bad_ld_p:
	nop
	mv_s	#$0bad001d,r0
	copy	r0,r1
	copy	r1,r2
	halt
	nop
	nop
	rts
	nop
	range ru

	;;
	;; subroutine:
	;; flush any pixels done (in the
	;; buffer pointed to by "writebuf")
	;; out to the screen
	;; This version is the simple
	;; one (no read-modify-write)
	;;
FlushOutput:
	cmp	#0,pixels_out
	rts	eq,nop		;; if no pixels to output, do nothing

	;; swap write and dma buffers
{	mv_s	writebuf,dmaoutbuf
	copy	dmaoutbuf,writebuf
}

	;; set up (xy) addressing for the new
	;; pixel write buffer
	st_io	writebuf,xybase
	st_io	#0,rx
	
	;; wait for main bus DMA to be
	;; finished
`waitdma:
	ld_io	mdmactl,temp0
	nop
	bits	#4,>>#0,temp0
	bra	ne,`waitdma,nop

	;; start a dma of the "writebuf"
	;; buffer

	;; register usage: we'll set up a
	;; small vector in dmavec, containing
	;; xcount, xposition, ycount, yposition
	
	;; check to see which direction we're
	;; stepping in

{	cmp	#0,y_inc		; check direction we're traversing in
	mv_s	#(1<<16),dmavec[0]	; initialize x count
}
	bra	ne,`step_in_y
{	mv_s	dmavec[0],dmavec[2]	; BRANCH DELAY SLOT: initialize y count
	copy	x_cur,dmavec[1]		; initialize current x
}
{	bra	`startdma		; BRANCH DELAY SLOT
	mv_s	y_cur,dmavec[3]		; BRANCH DELAY SLOT: initialize current y
}
	lsl	#16,pixels_out,dmavec[0] ; reset x count (only if branch to `step_in_y not taken
	add	pixels_out,>>#-16,x_cur	 ; and update x
`step_in_y:
	lsl	#16,pixels_out,dmavec[2] ; reset y count
	add	pixels_out,>>#-16,y_cur	 ; and update y
`startdma:
	ld_io	rc0,temp0		; pre-fetch rc0 (we'll need it later)
	st_sv	dmavec,dest_xptr	; set up x and y values for DMA	
	st_s	dmaoutbuf,dest_intaddr	; set up buffer address

	;;
	;; figure how many bytes the main code is allowed to
	;; write into our output buffer
	;; this is the smaller of rc0 and OUTSIZE
	;; we fetched rc0 into temp0 earlier
	st_io	#dest_cmd,mdmacptr	; start DMA command
{	cmp	#OUTSIZE,temp0
	st_io	temp0,rc1		; default to rc1 = temp0
}
	rts	lt			; if temp0 < OUTSIZE, use rc1 = temp0
	rts				; the return takes effect only if the earlier one doesn't
	mv_s	#0,pixels_out		; always executed
	st_io	#OUTSIZE,rc1		; happens only if the "rts lt" doesn't

	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;	FlushOutput with RMW
;;	reads the old data and merges it with the pixels, using
;;	the pixel alpha values
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	
FlushOutputRMW:
	cmp	#0,pixels_out
	rts	eq,nop

	;; swap write and dma buffers
{	mv_s	writebuf,dmaoutbuf
	copy	dmaoutbuf,writebuf
}
	
	st_io	#0,rx
	
	;; wait for main bus DMA to be
	;; finished
`waitdma:
	ld_io	mdmactl,temp0
	nop
	bits	#4,>>#0,temp0
	bra	ne,`waitdma,nop

	;; read in the old data
	;; register usage: we'll set up a
	;; small vector in dmavec, containing
	;; xcount, xposition, ycount, yposition
	
	;; check to see which direction we're
	;; stepping in

{	cmp	#0,y_inc		; check direction we're traversing in
	mv_s	#(1<<16),dmavec[0]	; initialize x count
}
	bra	ne,`step_in_y
{	mv_s	dmavec[0],dmavec[2]	; BRANCH DELAY SLOT: initialize y count
	copy	x_cur,dmavec[1]		; initialize current x
}
{	bra	`startdma		; BRANCH DELAY SLOT
	mv_s	y_cur,dmavec[3]		; BRANCH DELAY SLOT: initialize current y
}
	lsl	#16,pixels_out,dmavec[0] ; reset x count (only if branch to `step_in_y not taken
	add	pixels_out,>>#-16,x_cur	 ; and update x
`step_in_y:
	lsl	#16,pixels_out,dmavec[2] ; reset y count
	add	pixels_out,>>#-16,y_cur	 ; and update y
`startdma:
	ld_s	dest_flags,temp0	; get flags
	st_sv	dmavec,dest_xptr	; set up x and y values for DMA
{	bset	#13,temp0		; set the "read" flag
	st_s	#rmw_buf,dest_intaddr	; set up the scratch buffer address
}
	st_s	temp0,dest_flags	; set the "read" command
	st_io	#dest_cmd,mdmacptr	; start read of the old data

	
	;; wait for read to complete
`waitread:
	ld_io	mdmactl,temp0
	nop
	bits	#4,>>#0,temp0
	bra	ne,`waitread,nop

	;; now go through and merge the new pixels with the old
	mv_s	#rmw_buf,temp0
`rmw_mergeloop:
	ld_pz	(xy),v0			; fetch new pixel
{	ld_pz	(temp0),pixelvec	; fetch old pixel
	add	#4,temp0
}
	not	r3
	sub_p	pixelvec,v0		; v0 = new - old
	lsr	#2,r3			; convert alpha to 2.30
{	mul_p	r3,v0,>>#30,v0
	sub	#1,pixels_out		; decrement count of remaining pixels
}
	bra	gt,`rmw_mergeloop
	add_p	v0,pixelvec
{	st_p	pixelvec,(xy)		; store back merged new pixel
	addr	#1<<16,rx
}
	;; write back the modified pixels
	ld_s	dest_flags,r0
	st_io	#0,rx
	bclr	#13,r0			; clear the "read" flag
	st_s	r0,dest_flags
	st_s	dmaoutbuf,dest_intaddr
	ld_io	rc0,temp0		; pre-fetch rc0 (we'll need it later)

	;;
	;; figure how many bytes the main code is allowed to
	;; write into our output buffer
	;; this is the smaller of rc0 and OUTSIZE
	;; we fetched rc0 into temp0 earlier
	st_io	#dest_cmd,mdmacptr	; start DMA command
{	cmp	#OUTSIZE,temp0
	st_io	temp0,rc1		; default to rc1 = temp0
}
{	rts	lt			; if temp0 < OUTSIZE, use rc1 = temp0
	st_io	writebuf,xybase		; update (xy) addressing for new write buffer
}
	rts				; the return takes effect only if the earlier one doesn't
	mv_s	#0,pixels_out		; always executed
	st_io	#OUTSIZE,rc1		; happens only if the "rts lt" doesn't


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	;;
	;; DMA utility routines
	;;
	;; Copyright 1997 VM Labs, Inc.
	;; All rights reserved.
	;; This file is confidential and proprietary
	;; information of VM Labs disclosed pursuant
	;; to the non-disclosure agreement between
	;; VM Labs and the Recipient.
	;;
	;;

	; internal register usage
	dma_size = r0
	dma_extaddr = r1
	dma_intaddr = r2
	dma_ctlreg = r3
	dma_flags = r4
	xfer_size = r5
	cur_size = r6
	dma_temp0 = r7

	;
	; subroutine dma_write:
	; read some number of bytes from external RAM
	; into internal RAM
	;
	; parameters:
	; r0 == external address
	; r1 == internal address
	; r2 == number of bytes
	;
	; Registers modified:
	; v0, v1
	;
	; Stack used:
	; 4 long words (used as a DMA transfer buffer)
	;
h_dma_write:
	bra	common_dma	; next two instructions are in delay slots
	nop
	;; fall through to h_dma_read -- note that the
	;; first instruction of h_dma_read also gets executed
	;; by dma_write!!!


	;
	; subroutine h_dma_read:
	; read some number of bytes from external RAM
	; into internal RAM.
	;
	; parameters:
	; r0 == external address
	; r1 == internal address
	; r2 == number of bytes
	;
	; Registers modified:
	; v0, v1
	;
	; Stack used:
	; 4 long words (used as a DMA transfer buffer)
	;

	; Maximum size of one DMA chunk, in bytes.
	; Make this a multiple of 16 bytes, so
	; that other bus DMAs that start aligned
	; will finish aligned (helps avoid an
	; other bus DMA bug). Also, instruction
	; RAM writes must be vector multiples.
	;
	
MAX_SIZE = (60*4)
	
h_dma_read:
	;;;; NOTE NOTE NOTE: first instruction of h_dma_read
	;;;; is also executed by dma_write above (it is
	;;;; in a delay slot)
	mv_s	#0,dma_flags
	bset	#13,dma_flags		; set read bit

	; fall through

	;
	; Common subroutine, used for both
	; reads and writes
	;
	
common_dma:
	
	; round up to next largest number of vectors
	; (WARNING: if the transfer isn't aligned, the other bus
	; may have problems -- it's the caller's responsibility
	; to look out for this!)
{	add	#ALIGN,r2,xfer_size
	mv_s	r1,dma_intaddr
}
	and	#~ALIGN,xfer_size
{	copy	r0,dma_extaddr
	mv_s	#addrof(mdmactl),dma_ctlreg
}
	
	push	v0		; create a 16 byte buffer on the stack
	
	; check for which bus to use
	; our rule of thumb: if the upper nybble is 4, use the
	; main bus, otherwise use the other bus
	; (the other bus is better for MPE<->MPE transfers,
	; because the main bus is buggy)
	;
	copy	dma_extaddr,dma_temp0
	bits	#3,>>#28,dma_temp0		; extract top nybble
	cmp	#4,dma_temp0
	bra	eq,start_xfer,nop
	sub	#addrof(mdmactl)-addrof(odmactl),dma_ctlreg   ;; set dma_ctlreg to addrof(odmactl)

start_xfer:
	;
	; Figure out how many bytes to DMA this time
	; around the loop; do at most MAX_SIZE bytes
	;
	
	mv_s	#MAX_SIZE,cur_size
	cmp	cur_size,xfer_size
	bra	ge,bigxfer,nop		; if (xfer_size - max_size >= 0, use max_size)	
	mv_s	xfer_size,cur_size

bigxfer:
	asl	#(16-2),cur_size,dma_size	; put in the upper 16 bits, as words

	; wait for DMA pending to go clear
wait_pending:
	ld_s	(dma_ctlreg),dma_temp0
	or	dma_flags,dma_size	; now DMA flags are set up for transfer
					;; (it's OK to do the "or" multiple times)
{	btst	#4,dma_temp0
	ld_io	sp,dma_temp0		; get current stack pointer (used as DMA buffer)
}
	bra	ne,wait_pending,nop

	
{	st_v	v0,(dma_temp0)
	add	#16,dma_ctlreg		; point to control register
}
	; launch the DMA
	st_s	dma_temp0,(dma_ctlreg)	; start DMA
	sub	#16,dma_ctlreg

	; update DMA variables
	sub	cur_size,xfer_size
	bra	gt, start_xfer
	add	cur_size,dma_extaddr
	add	cur_size,dma_intaddr
	
	; wait for the DMA to finish
wait_dma:
	ld_s	(dma_ctlreg),dma_temp0
	nop
	bits	#4,>>#0,dma_temp0
	bra	ne,wait_dma,nop

	; restore the stack pointer with pop v0
	; NOTE: this does *not* restore the contents of
	; v0, because we were using the stack
	; as a buffer for the DMA command
	
	pop	v0	
	rts
	nop
	nop

	.if defined(USE_MINI_BIOS)
	.segment blitd
	.else
	.segment mblitd
	.endif
	
	.origin 0x20100000

	;; the CLUT must be on a 256*4 byte boundary
	.align 256*4
CLUT:
	;; initialize the CLUT to contain a grey scale
	i = 0
	.while i < 256
	.dc.s	(i << 24) | 0x00808000
	i = i+1
	.end

	;;
	;; source DMA command
	;;
	
	.align.v
src_cmd:
src_flags:
	.dc.s	0
src_baseaddr:
	.dc.s	0
src_xptr:
	.dc.s	0
src_yptr:
	.dc.s	0
src_intaddr:
	.dc.s	0
	
linear_flag:
	.dc.s	0		;; flag for linear/non-linear input
src_width:
	.dc.s	0		;; width of linear bitmap

	;;
	;; storage space for a linear DMA command
	;;
	.align.v
lin_dma_cmd:
	.dc.s	0,0,0,0
	
	;;
	;; destination DMA command
	;;
	.align.v
dest_cmd:
dest_flags:
	.dc.s	((352/8)<<16) | (6<<13) | (4<<4) | (1<<11)
dest_baseaddr:
	.dc.s	0x40000000	; address of frame buffer
dest_xptr:
	.dc.s	0		; X pointer and length
dest_yptr:
	.dc.s	0		; Y pointer and length
dest_intaddr:
	.dc.s	0		; internal address
	

	.align.v
DEBUG_LAST_CMD:
	.dc.s	0,0,0,0
img_alphas:
	.dc.s	0		; flag for whether or not to use alpha
	
shading_factor:
	.dc.s	0		; 2.30 lighting factor
	
	;; allow for up to 4 extra words to be read
	;; then allow for vector alignment requirements on system bus RAM
	
	.align.v
inp_buf:
	.ds.s	INPSIZE + 4

	.align.v
out_buf1:	
	.ds.s	OUTSIZE	
out_buf2:
	.ds.s	OUTSIZE
rmw_buf:
	.ds.s	OUTSIZE
		
.if 0
LUMASCALE=220/256
CHROMASCALE=224/256
.else
LUMASCALE=200/256
CHROMASCALE=200/256
.endif
	
	.align.v
rgbmatrix:
	.dc.s	fix(LUMASCALE*0.299,30), fix(LUMASCALE*0.587,30), fix(LUMASCALE*0.114,30), 0
	.dc.s	fix(CHROMASCALE*0.500,30), fix(CHROMASCALE*-0.419,30), fix(CHROMASCALE*-0.081,30), 0
	.dc.s   fix(CHROMASCALE*-0.169,30), fix(CHROMASCALE*-0.331,30), fix(CHROMASCALE*0.500,30), 0


	.align.v
last_request_mpe:
	.dc.s	0		; last MPE to send us a request

