// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id: i_system.c,v 1.1.1.1 1999/12/20 14:23:41 ersmith Exp $
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log: i_system.c,v $
// Revision 1.1.1.1  1999/12/20 14:23:41  ersmith
// Initial import of the VMLabs version of Doom
//
//
// DESCRIPTION:
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: i_system.c,v 1.1.1.1 1999/12/20 14:23:41 ersmith Exp $";


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <stdarg.h>
#include <sys/time.h>
#include <unistd.h>

#include "doomdef.h"
#include "m_misc.h"
#include "i_video.h"
#include "i_sound.h"

#include "d_net.h"
#include "g_game.h"

#ifdef __GNUG__
#pragma implementation "i_system.h"
#endif
#include "i_system.h"


#ifdef MERLIN
#include <nuon/mutil.h>
#include <nuon/dma.h>
#include <nuon/sdram.h>
#endif


int	mb_used = 4;


void
I_Tactile
( int	on,
  int	off,
  int	total )
{
  // UNUSED.
  on = off = total = 0;
}

ticcmd_t	emptycmd;
ticcmd_t*	I_BaseTiccmd(void)
{
    return &emptycmd;
}


int  I_GetHeapSize (void)
{
    return mb_used*1024*1024;
}

byte* I_ZoneBase (int*	size)
{
    *size = mb_used*1024*1024;
#ifdef MERLIN_USE_SDRAM
    return (byte *)SDRAMAlloc(*size);
#else
    return (byte *) malloc (*size);
#endif
}



//
// I_GetTime
// returns time in 1/70th second tics
//
int  I_GetTime (void)
{
    static int		basetime=0;
    int			newtics;
#ifdef MERLIN_OLD
     if (!basetime) {
	InitTimer();
	basetime = 1;
    }
    newtics = GetTimer((void *)0, (void *)0);
//     newtics = times((void *)0);
    return newtics*TICRATE/1000;
#else
    struct timeval	tp;
    struct timezone	tzp;

    gettimeofday(&tp, &tzp);
    if (!basetime)
	basetime = tp.tv_sec;
    newtics = (tp.tv_sec-basetime)*TICRATE + tp.tv_usec*TICRATE/1000000;
    return newtics;
#endif
}



//
// I_Init
//
void I_Init (void)
{
    I_InitSound();
    //  I_InitGraphics();
}

//
// I_Quit
//
void I_Quit (void)
{
    D_QuitNetGame ();
    I_ShutdownSound();
    I_ShutdownMusic();
    M_SaveDefaults ();
    I_ShutdownGraphics();
    exit(0);
}

void I_WaitVBL(int count)
{
#ifdef MERLIN
#include <nuon/mutil.h>
    VidSync();
#else
#ifdef SGI
    sginap(1);                                           
#else
#ifdef SUN
    sleep(0);
#else
    usleep (count * (1000000/70) );                                
#endif
#endif
#endif
}

void I_BeginRead(void)
{
}

void I_EndRead(void)
{
}

byte*	I_AllocLow(int length)
{
    byte*	mem;
#ifdef MERLIN
    mem = (byte *)SDRAMAlloc((unsigned long)length);
#else
    mem = (byte *)malloc (length);
#endif
    memset (mem,0,length);
    return mem;
}

void kprintf (const char fmt[], ...);

//
// I_Error
//+
extern boolean demorecording;
char debugStr[1024];

void I_Error (char *error, ...)
{
    va_list	argptr;

    // Message first.
    va_start (argptr,error);
#if 1
    fprintf (stderr, "Error: ");
    vfprintf (stderr,error,argptr);
    fprintf (stderr, "\n");
#else
    kprintf ("Error: ");
    vsprintf (debugStr,error,argptr);
    kprintf ("%s\n",debugStr);
#endif
    va_end (argptr);

    fflush( stderr );

    // Shutdown. Here might be other errors.
    if (demorecording)
	G_CheckDemoStatus();

    D_QuitNetGame ();
    I_ShutdownGraphics();

	for(;;);    
    exit(-1);
}
